list(
  # Adjustment name ----
  Name = "Multiple Imputations (mice)",

  # Adjustment type ----
  Type = "MULTIPLE_IMPUTATIONS",

  # Adjustment subtype ----
  SubType = "MICE",

  # Input parameters to the adjustment function ----
  Parameters = list(
    # Parameter 1: a specification with label, type and default value
    nimp = list(
      label = "Number of imputations",
      value = 2L,
      input = "numeric"),
    # Parameter 2
    nit = list(
      label = "Number of mice iterations",
      value = 10L,
      input = "numeric"),
    # Parameter 3
    nsdf = list(
      label = "Number of degrees of freedom for spline",
      value = 6L,
      input = "numeric"),
    # Parameter 5
    runInParallel = list(
      label = "Run in parallel",
      value = FALSE,
      input = "checkbox")
  ),

  # Names of packages that must be made available to the adjustment function ----
  RequiredPackageNames = c("mice"),

  ## Adjustment function ----
  AdjustmentFunction = function(inputData, parameters) {

    # Perform imputations per data set.
    # This is the actual worker function.
    workerFunction <- function(dataSet, nit, nimp, nsdf) {

      # Define covariates
      xColNames <- c("AIDS")
      # Define outcomes
      yColNames <- c("Age", "SqCD4", "Transmission", "Migr")

      # Create splines with proper names and intercept
      splineBasisMatrix <- as.data.table(splines::ns(dataSet$DY, df = nsdf))
      setnames(splineBasisMatrix, paste0("V", colnames(splineBasisMatrix)))
      intercept <- 1L

      # Define covariates of joint imputation model
      X <- cbind(dataSet[, ..xColNames],
                 Intercept = intercept,
                 splineBasisMatrix)
      # Define outcomes of joint imputation model
      Y <- dataSet[, ..yColNames]

      # Workaround for jomo bug:
      # Not used levels must be removed
      X <- droplevels(X)
      Y <- droplevels(Y)

      # Run model
      imp <- mice::mice(cbind(Y, X),
                        m = nimp,
                        maxit = nit)
      imp <- setDT(mice::complete(imp, action = "long", include = TRUE))
      setnames(imp, old = c(".imp", ".id"), new = c("Imputation", "id"))

      indexColNames <- c("Imputation", "id")
      impColNames <- union(indexColNames, yColNames)
      dataSetColNames <- setdiff(colnames(dataSet), yColNames)

      mi <- cbind(imp[, ..impColNames], dataSet[, ..dataSetColNames])

      setcolorder(mi, union(indexColNames, names(dataSet)))

      ConvertDataTableColumns(mi, c(Imputation = function(x) as.integer(as.character(x))))

      return(mi)
    }

    # 1. Preprocess data for mice specifically
    inputData <- PreProcessInputDataForMice(inputData)

    # 2. Save original order for later
    inputData[, OrigSort := .I]

    # 3. Measures years from earlier diagnosis year
    inputData[, DY := DateOfDiagnosisYear - min(DateOfDiagnosisYear)]

    # 4. Split by gender to data sets
    dataSets <- split(inputData, by = c("Gender"))

    # 5. Execute the worker function per data set
    if (parameters$runInParallel) {
      # Run in parallel
      cl <- parallel::makeCluster(2)
      parallel::clusterExport(cl, "libPaths")
      parallel::clusterEvalQ(cl, {
				.libPaths(libPaths)
        library(data.table)
      })
      outputData <- parallel::parLapply(cl,
                                        dataSets,
                                        workerFunction,
                                        nit = parameters$nit,
                                        nimp = parameters$nimp,
                                        nsdf = parameters$nsdf)
      parallel::stopCluster(cl)
    } else {
      # Run sequentially
      outputData <- lapply(dataSets,
                           workerFunction,
                           nit = parameters$nit,
                           nimp = parameters$nimp,
                           nsdf = parameters$nsdf)
    }

    # 6. Combine all data sets
    outputData <- rbindlist(outputData)

    # 7. Restore original order per Imputation
    setorder(outputData, Imputation, OrigSort)

    return(
      list(Table = outputData,
           Artifacts = list())
    )
  }
)
